/*
 * _SideChannelAttack_01.xc
 *
 *  Created on: 1. mai 2020
 *      Author: teig
 *      Ver 0.1.9  12May2020 using code from _xc_race_5_2020
 *      Ver 0.1.8            Made a project _xc_race_5_2020 from here, that becase XMOS Ticket
 *                           [Product Bug #32552] AutoReply: Possible race by timerafter case over interface input case in selects in equal code on 7 cores
 *      Ver 0.1.7            DEBUG_PRINT_TEST 1 race
 *      Ver 0.1.6            inSW1SW2_buttons is new
 *      Ver 0.1.5            XS1_PORT_1M exterranl black switch
 *                           timeout_ticks = 0; // stop_jamming never reached
 *      Ver 0.1.4            jamming_task_start_stop probably better, here with log results
 *      Ver 0.1.3            jamming_task_start_stop is not optimal, it will not handle 0 (no ones) and -1 (all ones) ok
 *      Ver 0.1.2  05Feb2020 RGB_RED also used
 *      Ver 0.1.1  05Feb2020 Initial with button start
 */

#include <platform.h> // core
#include <stdio.h>    // printf
#include <timer.h>    // delay_milliseconds(..), XS1_TIMER_HZ etc
#include <iso646.h>   // readability
#include <stdint.h>   // uint32_t

// ---
// Control printing
// See https://stackoverflow.com/questions/1644868/define-macro-for-debug-printing-in-c
// ---

#define DEBUG_PRINT_TEST 1
#define debug_print(fmt, ...) do { if(DEBUG_PRINT_TEST) printf(fmt, __VA_ARGS__); } while (0)

#define DEBUG_PRINT_TEST_X 1
#define debug_print_x(fmt, ...) do { if(DEBUG_PRINT_TEST_X) printf(fmt, __VA_ARGS__); } while (0)

// BOOLEAN #include <stdbool.h> if C99
// See http://www.teigfam.net/oyvind/home/technology/165-xc-code-examples/#bool
typedef enum {false,true} bool; // 0,1 This typedef matches any integer-type type like long, int, unsigned, char, bool

#define BITS32_LEN 32

#define t_swap(type,a,b) {type t = a; a = b; b = t;}

typedef signed int time32_t; // signed int (=signed) or unsigned int (=unsigned) both ok, as long as they are monotoneously increasing
                             // XC/XMOS 100 MHz increment every 10 ns for max 2exp32 = 4294967296,
                             // ie. divide by 100 mill = 42.9.. seconds

// ---
// 4 BIT PORT
// Built in switches SW1 and SW2
// --
in buffered port:4 inSW1SW2_buttons = on tile[0]: XS1_PORT_4E; // SW1 and SW2 on the xCORE-200 eXplorer board
//
#define BOARD_SW1_MASK 0x01
#define BOARD_SW2_MASK 0x02

//
#define BUTTON_PRESSED  0
#define BUTTON_RELEASED 1

// ---
// 4 BIT PORT
// Internal LEDs defined. High is "on"
// ---
out buffered port:4 outP4_leds = on tile[0]: XS1_PORT_4F; // xCORE-200 explorerKIT GPIO J1 P5, P3, P1
//
#define BOARD_LEDS_INIT           0x00
#define BOARD_LED_MASK_GREEN_ONLY 0x01 // BIT0
#define BOARD_LED_MASK_RGB_BLUE   0x02 // BIT1
#define BOARD_LED_MASK_RGB_GREEN  0x04 // BIT2
#define BOARD_LED_MASK_RGB_RED    0x08 // BIT3

#define NUM_JAMMING_SERVERS 7
//
typedef interface jamming_if {
    void start_jamming (const unsigned bit_pattern); // Using timer ticks as bit_pattern is just first approximation
    void stop_jamming (void);
} jamming_if;

#define MY_KEY_NUM_CHARS 16

typedef interface encrypt_key_if {
    void encrypt (char key [], static const size_t len, const unsigned bit_pattern);
} encrypt_key_if;

#define STEP_LEN_MICROSECS 1

/*
 * FUNCTION rotate_32_left_by_1
 *
 * Rotates uint32_t by one bit left
 * Inspired by https://en.wikipedia.org/wiki/Circular_shift
 */
uint32_t rotate_32_left_by_1 (const uint32_t value) {

    // value shifted left by one pos. and pos0 comes in as 0
    // bitor
    // bit32 into pos0 where only pos0 is or'ed in

    return ((value<<1) bitor ((value>>31) bitand 1));
}

/*
 * TASK jamming_task
 * Taken from jamming_task_2 in _xc_race_5_2020 12May2020
 */
void jamming_task (const unsigned iof_task, server jamming_if i_jamming) {

    timer    timeout_tmr;
    time32_t timeout_ticks;
    unsigned timeout_cnt = 0;

    timer    immediate_tmr;
    time32_t immediate_ticks;
    unsigned immediate_cnt = 0;

    time32_t bit_pattern_in;
    bool timeout_do = false;
    bool immediate_do = false;
    unsigned iof_step = 0;

    while (1) {
        select {
            case i_jamming.start_jamming (const unsigned bit_pattern): {
                debug_print ("task_2 %02u start_jamming bit_pattern %08X\n", iof_task, bit_pattern);
                bit_pattern_in = bit_pattern;
                immediate_do = true;
                immediate_tmr :> immediate_ticks; // Immediately
            } break;

            case i_jamming.stop_jamming (void): { // qwe This is never taken if race appears
                debug_print ("task_2 %02u stop_jamming imm=%u tim=%u\n", iof_task, immediate_cnt, timeout_cnt);
                timeout_do = false;
                immediate_do = false;
                immediate_cnt = 0;
                timeout_cnt = 0;
            } break;

            case (immediate_do) => immediate_tmr when timerafter (immediate_ticks) :> void : {
                if ((bit_pattern_in bitand (1<<iof_step))==0) {
                    immediate_cnt++;
                    debug_print ("task_2 %02u I run %u\n", iof_task, immediate_cnt);
                    immediate_tmr :> immediate_ticks;
                } else { // bit is 1
                    // SWITCH TO TIMEOUT
                    timeout_cnt++;
                    debug_print ("task_2 %02u I wait %u\n", iof_task, timeout_cnt);
                    immediate_do = false;
                    timeout_do = true;
                    timeout_tmr :> timeout_ticks;
                    timeout_ticks += (STEP_LEN_MICROSECS * XS1_TIMER_MHZ);
                }
                iof_step = (iof_step + 1) % BITS32_LEN; // [0..31]
            } break;

            case (timeout_do) => timeout_tmr when timerafter (timeout_ticks) :> void : {
                if ((bit_pattern_in bitand (1<<iof_step))==0) { // bit is 0
                    // SWITCH TO IMMEDIATE
                    immediate_cnt++;
                    debug_print ("task_2 %02u T run %u\n", iof_task, immediate_cnt);
                    timeout_do = false;
                    immediate_do = true;
                    immediate_tmr :> immediate_ticks;
                } else { // bit is 1
                    timeout_cnt++;
                    debug_print ("task_2 %02u T wait %u\n", iof_task, timeout_cnt);
                    timeout_ticks += (STEP_LEN_MICROSECS * XS1_TIMER_MHZ);
                }
                iof_step = (iof_step + 1) % BITS32_LEN; // [0..31]
            } break;
        }
    }
}

/*
 * TASK jamming_task_start_stop
 * Taken from jamming_task_start_stop in _xc_race_5_2020 12May2020
 */
void jamming_task_start_stop(const bool start_stop, // if "false" then..
        const unsigned &?bit_pattern,               // ..this is "null", therefore "if isnull(timeout_ticks)".. not needed
        client jamming_if i_jamming[NUM_JAMMING_SERVERS]) {

    unsigned bit_pattern_copy;

    if (start_stop == true) { // start
        unsigned invert_bitpos = bit_pattern bitand (BITS32_LEN-1); // Invert one bitpos decided by 5 least significant bits

        debug_print ("invert_bit_pos %u\n", invert_bitpos);

        bit_pattern_copy = bit_pattern;

        #if (RACE_DO_XOR==1)
            bit_pattern_copy xor_eq (1 << invert_bitpos); // One or many bits inverted makes little difference. But one pos will do "sliding doors" and not create any zeros
            // I could not use a psudo-random sequence here, because it may produce 0x00000000
        #else
            // 0 means don't invert bit
        #endif

        for (int i_jamming_server = 0; i_jamming_server < NUM_JAMMING_SERVERS; i_jamming_server++) {

            if (bit_pattern_copy == 0) {
                debug_print("%08X bit_pattern ZERO\n", bit_pattern_copy);
            } else {
                debug_print("%08X bit_pattern\n", bit_pattern_copy);
            }
            i_jamming[i_jamming_server].start_jamming(bit_pattern_copy);

            bit_pattern_copy = rotate_32_left_by_1(bit_pattern_copy);
        }
    } else { // stop
        for (int i_jamming_server = 0; i_jamming_server < NUM_JAMMING_SERVERS;
                i_jamming_server++) {
            i_jamming[i_jamming_server].stop_jamming();
        }
    }
}
/*
 * TASK encryption_task
 * Taken from encryption_task in _xc_race_5_2020 12May2020
 */
[[combinable]]
void encryption_task(server encrypt_key_if i_encrypt,
        client jamming_if i_jamming[NUM_JAMMING_SERVERS]) {

    while (1) {
        select {
            case i_encrypt.encrypt (char key [], static const size_t len, const unsigned bit_pattern) : {
                unsigned sum = 0;

                debug_print_x ("%s", "starting\n");
                jamming_task_start_stop (true, bit_pattern, i_jamming);
                debug_print_x ("%s", "started\n");

                // Do something here

                debug_print_x ("%s", "stopping\n");
                jamming_task_start_stop (false, null, i_jamming);// I like "null" when a parameter is not used
                debug_print_x ("%s", "SUCCESS: STOPPED!\n\n"); // If RACE_NO_STOP_JAMMING this is never reached

            } break;
        }
    }
}

/*
 * KEY. TO BE KNOWN BY EXTERNAL SPY, TO KNWO WHAT TO SEARCH FPR
 */
#define MY_KEY "SCA-1234567abcd" // The idea is the the expernal spy should KNOW this "key"
//             "123456789012345" // MY_KEY_NUM_CHARS-1 = 15 chars, the 16th is the nul-terminator added by the compiler, which is now part of the key

/*
 * encryption_task_starter
 * [[combinable]] = may share a logical core
 */
[[combinable]]
void encryption_task_starter (
        in buffered port:4    inSW1SW2_buttons,
        out buffered port:4   outP4_leds,
        client encrypt_key_if i_encrypt) {

    timer    tmr;
    time32_t timeout_ticks;
    char     key []              = MY_KEY;
    bool     buttons_current_val = BUTTON_RELEASED;
    unsigned led_bits            = BOARD_LEDS_INIT; // All off

    outP4_leds <: led_bits;

    while (1) {
        select {
            case inSW1SW2_buttons when pinsneq(buttons_current_val) :> buttons_current_val: {
                // I/O pin changed value
                // Real debouncing not done (best done in separate task, with its own timerafter)

                if ((buttons_current_val bitand BOARD_SW2_MASK) == BUTTON_PRESSED) {

                    led_bits or_eq (BOARD_LED_MASK_GREEN_ONLY bitor BOARD_LED_MASK_RGB_RED);
                    outP4_leds <: led_bits; // GREEN_ONLY and RGB_RED on, seen on the outside, could be used to trigger scope

                    // I think that this obfuscation should not be known by external spy (1)
                    //
                    tmr :> timeout_ticks; // Closest to real random value I could think of. Nice bit_pattern

                    debug_print_x ("%08X bit_pattern\n", timeout_ticks);
                    i_encrypt.encrypt (key, sizeof (key), (unsigned) timeout_ticks);

                    led_bits and_eq (compl BOARD_LED_MASK_RGB_RED);
                    outP4_leds <: led_bits; // RGB_RED OFF

                    delay_milliseconds(100); // To see the GREEN_ONLY
                    led_bits and_eq (compl BOARD_LED_MASK_GREEN_ONLY);
                    outP4_leds <: led_bits; // GREEN_ONLY OFF
                } else {
                    delay_milliseconds(100); // Rude debounce works rather ok
                }
            } break;
        }
    }
}

/*
 * main
 */
int main(void) {
    interface encrypt_key_if i_encrypt;
    interface jamming_if     i_jamming[NUM_JAMMING_SERVERS];

    par {
        [[combine]]
        par {
            encryption_task_starter (inSW1SW2_buttons, outP4_leds, i_encrypt);
            encryption_task         (i_encrypt, i_jamming); // external spy should DISCOVER "key" from internal calculations here
        }

        // The idea is that while the encryption is being done then these tasks should each run
        // on a separate core alone. They could do other stuff when not doing this, but they
        // are not coded with any such extra stuff
        //
        par (int iOf_server = 0; iOf_server < NUM_JAMMING_SERVERS; iOf_server++) {
            jamming_task (iOf_server, i_jamming[iOf_server]);
        }
    }
    return 0;
}
